searchNodes=[{"type":"behaviour","title":"Spark.Dsl","doc":"The primary entry point for adding a DSL to a module. To add a DSL to a module, add use Spark.Dsl, ...options . The options supported with use Spark.Dsl are: :single_extension_kinds (list of atom/0 ) - The extension kinds that are allowed to have a single value. For example: [:data_layer] The default value is [] . :many_extension_kinds (list of atom/0 ) - The extension kinds that can have multiple values. e.g [notifiers: [Notifier1, Notifier2]] The default value is [] . :untyped_extensions? ( boolean/0 ) - Whether or not to support an extensions key which contains untyped extensions The default value is true . :default_extensions ( keyword/0 ) - The extensions that are included by default. e.g [data_layer: Default, notifiers: [Notifier1]] Default values for single extension kinds are overwritten if specified by the implementor, while many extension kinds are appended to if specified by the implementor. The default value is [] . :opt_schema ( keyword/0 ) - A schema for additional options to accept when calling use YourSpark The default value is [] . See the callbacks defined in this module to augment the behavior/compilation of the module getting a Dsl. Schemas/Data Types Spark DSLs use a superset of NimbleOptions for the schema that makes up sections/entities of the DSL. For more information, see Spark.OptionsHelpers .","ref":"Spark.Dsl.html"},{"type":"callback","title":"Spark.Dsl.explain/2","doc":"Validate/add options. Those options will be passed to handle_opts and handle_before_compile","ref":"Spark.Dsl.html#c:explain/2"},{"type":"callback","title":"Spark.Dsl.handle_before_compile/1","doc":"Handle options in the context of the module, after all extensions have been processed. Must return a quote block.","ref":"Spark.Dsl.html#c:handle_before_compile/1"},{"type":"function","title":"Spark.Dsl.handle_fragments/2","doc":"","ref":"Spark.Dsl.html#handle_fragments/2"},{"type":"callback","title":"Spark.Dsl.handle_opts/1","doc":"Handle options in the context of the module. Must return a quote block. If you want to persist anything in the DSL persistence layer, use @persist {:key, value} . It can be called multiple times to persist multiple times.","ref":"Spark.Dsl.html#c:handle_opts/1"},{"type":"callback","title":"Spark.Dsl.init/1","doc":"Validate/add options. Those options will be passed to handle_opts and handle_before_compile","ref":"Spark.Dsl.html#c:init/1"},{"type":"function","title":"Spark.Dsl.is?/2","doc":"","ref":"Spark.Dsl.html#is?/2"},{"type":"callback","title":"Spark.Dsl.verify/2","doc":"A callback that is called in the after_verify hook. Only runs on versions of Elixir &gt;= 1.14.0","ref":"Spark.Dsl.html#c:verify/2"},{"type":"type","title":"Spark.Dsl.entity/0","doc":"","ref":"Spark.Dsl.html#t:entity/0"},{"type":"type","title":"Spark.Dsl.opts/0","doc":"","ref":"Spark.Dsl.html#t:opts/0"},{"type":"type","title":"Spark.Dsl.section/0","doc":"","ref":"Spark.Dsl.html#t:section/0"},{"type":"type","title":"Spark.Dsl.t/0","doc":"","ref":"Spark.Dsl.html#t:t/0"},{"type":"module","title":"Spark.Dsl.Builder","doc":"Utilities for building DSL objects programatically, generally used in transformers.","ref":"Spark.Dsl.Builder.html"},{"type":"macro","title":"Spark.Dsl.Builder.defbuilder/2","doc":"","ref":"Spark.Dsl.Builder.html#defbuilder/2"},{"type":"macro","title":"Spark.Dsl.Builder.defbuilderp/2","doc":"","ref":"Spark.Dsl.Builder.html#defbuilderp/2"},{"type":"function","title":"Spark.Dsl.Builder.handle_nested_builders/2","doc":"Handles nested values that may be {:ok, result} or {:error, term} , returning any errors and unwrapping any ok values This allows users of builders to do things like: dsl_state |&gt; Ash.Resource.Builder . add_new_action ( :update , :publish , changes : [ Ash.Resource.Builder . build_action_change ( Ash.Resource.Change.Builtins . set_attribute ( :state , :published ) ) ] ) If your builder function calls handle_nested_builders/2 with their input before building the thing its building.","ref":"Spark.Dsl.Builder.html#handle_nested_builders/2"},{"type":"type","title":"Spark.Dsl.Builder.input/0","doc":"","ref":"Spark.Dsl.Builder.html#t:input/0"},{"type":"type","title":"Spark.Dsl.Builder.result/0","doc":"","ref":"Spark.Dsl.Builder.html#t:result/0"},{"type":"module","title":"Spark.Dsl.Entity","doc":"Declares a DSL entity. A dsl entity represents a dsl constructor who's resulting value is a struct. This lets the user create complex objects with arbitrary(mostly) validation rules. The lifecycle of creating entities is complex, happening as Elixir is compiling the modules in question. Some of the patterns around validating/transforming entities have not yet solidified. If you aren't careful and don't follow the guidelines listed here, you can have subtle and strange bugs during compilation. Anything not isolated to simple value validations should be done in transformers . See Spark.Dsl.Transformer . An entity has a target indicating which struct will ultimately be built. An entity also has a schema . This schema is used for documentation, and the options are validated against it before continuing on with the DSL. To create positional arguments to the builder, use args . The values provided to args need to be in the provided schema as well. They will be positional arguments in the same order that they are provided in the args key. auto_set_fields will set the provided values into the produced struct (they do not need to be included in the schema). transform is a function that takes a created struct and can alter it. This happens immediately after handling the DSL options, and can be useful for setting field values on a struct based on other values in that struct. If you need things that aren't contained in that struct, use an Spark.Dsl.Transformer . This function returns {:ok, new_entity} or {:error, error} , so this can also be used to validate the entity. entities allows you to specify a keyword list of nested entities. Nested entities are stored on the struct in the corresponding key, and are used in the same way entities are otherwise. singleton_entity_keys specifies a set of entity keys (specified above) that should only have a single value. This will be validated and unwrapped into nil | single_value on success. identifier expresses that a given entity is unique by that field, validated by the DSL. Example @my_entity % Spark.Dsl.Entity { name : :my_entity , target : MyStruct , schema : [ my_field : [ type : :atom , required : false ] ] } Once compiled by Spark, entities can be invoked with a keyword list: my_entity my_field : :value Or with a do block: my_entity do my_field :value end For a full example, see Spark.Dsl.Extension .","ref":"Spark.Dsl.Entity.html"},{"type":"type","title":"Spark.Dsl.Entity.args/0","doc":"Specifies positional arguments for an Entity. An entity declared like this: @entity % Spark.Dsl.Entity { name : :entity , target : Entity , schema : [ positional : [ type : :atom , required : true ] , other : [ type : :atom , required : false ] , ] , args : [ :positional ] } Can be instantiated like this: entity :positional_argument do other :other_argument end","ref":"Spark.Dsl.Entity.html#t:args/0"},{"type":"type","title":"Spark.Dsl.Entity.auto_set_fields/0","doc":"Set the provided key value pairs in the produced struct. These fields do not need to be included in the Entity's schema.","ref":"Spark.Dsl.Entity.html#t:auto_set_fields/0"},{"type":"type","title":"Spark.Dsl.Entity.deprecations/0","doc":"","ref":"Spark.Dsl.Entity.html#t:deprecations/0"},{"type":"type","title":"Spark.Dsl.Entity.describe/0","doc":"User provided documentation. Documentation provided in a Entity 's describe field will be included by Spark in any generated documentation that includes the Entity .","ref":"Spark.Dsl.Entity.html#t:describe/0"},{"type":"type","title":"Spark.Dsl.Entity.docs/0","doc":"Internal field. Not set by user.","ref":"Spark.Dsl.Entity.html#t:docs/0"},{"type":"type","title":"Spark.Dsl.Entity.entities/0","doc":"A keyword list of nested entities.","ref":"Spark.Dsl.Entity.html#t:entities/0"},{"type":"type","title":"Spark.Dsl.Entity.examples/0","doc":"","ref":"Spark.Dsl.Entity.html#t:examples/0"},{"type":"type","title":"Spark.Dsl.Entity.hide/0","doc":"","ref":"Spark.Dsl.Entity.html#t:hide/0"},{"type":"type","title":"Spark.Dsl.Entity.id/0","doc":"","ref":"Spark.Dsl.Entity.html#t:id/0"},{"type":"type","title":"Spark.Dsl.Entity.imports/0","doc":"","ref":"Spark.Dsl.Entity.html#t:imports/0"},{"type":"type","title":"Spark.Dsl.Entity.links/0","doc":"","ref":"Spark.Dsl.Entity.html#t:links/0"},{"type":"type","title":"Spark.Dsl.Entity.modules/0","doc":"","ref":"Spark.Dsl.Entity.html#t:modules/0"},{"type":"type","title":"Spark.Dsl.Entity.name/0","doc":"","ref":"Spark.Dsl.Entity.html#t:name/0"},{"type":"type","title":"Spark.Dsl.Entity.no_depend_modules/0","doc":"","ref":"Spark.Dsl.Entity.html#t:no_depend_modules/0"},{"type":"type","title":"Spark.Dsl.Entity.recursive_as/0","doc":"","ref":"Spark.Dsl.Entity.html#t:recursive_as/0"},{"type":"type","title":"Spark.Dsl.Entity.singleton_entity_keys/0","doc":"","ref":"Spark.Dsl.Entity.html#t:singleton_entity_keys/0"},{"type":"type","title":"Spark.Dsl.Entity.snippet/0","doc":"","ref":"Spark.Dsl.Entity.html#t:snippet/0"},{"type":"type","title":"Spark.Dsl.Entity.t/0","doc":"","ref":"Spark.Dsl.Entity.html#t:t/0"},{"type":"type","title":"Spark.Dsl.Entity.target/0","doc":"Defines the struct that will be built from this entity definition. The struct will need to have fields for all entities , t:schema/0 fields, and auto_set_fields/0 .","ref":"Spark.Dsl.Entity.html#t:target/0"},{"type":"type","title":"Spark.Dsl.Entity.transform/0","doc":"Specifies a function that will run on the target struct after building. @my_entity % Spark.Dsl.Entity { name : :my_entity , target : MyEntity , schema : [ my_field : [ type : :list , required : true ] ] , transform : { MyModule , :max_three_items , [ ] } } def max_three_items ( my_entity ) do if length ( my_entity . my_field ) &gt; 3 do { :error , &quot;Can&#39;t have more than three items&quot; } else { :ok , my_entity } end end","ref":"Spark.Dsl.Entity.html#t:transform/0"},{"type":"behaviour","title":"Spark.Dsl.Extension","doc":"An extension to the Spark DSL. This allows configuring custom DSL components, whose configurations can then be read back. This guide is still a work in progress, but should serve as a decent example of what is possible. Open issues on Github if you have any issues/something is unclear. The example at the bottom shows how you might build a (not very contextually relevant) DSL extension that would be used like so: defmodule MyApp.Vehicle do use Spark.Dsl end defmodule MyApp.MyResource do use MyApp.Vehicle , extensions : [ MyApp.CarExtension ] cars do car :ford , :focus , trim : :sedan car :toyota , :corolla end end The extension: defmodule MyApp.CarExtension do @car_schema [ make : [ type : :atom , required : true , doc : &quot;The make of the car&quot; ] , model : [ type : :atom , required : true , doc : &quot;The model of the car&quot; ] , type : [ type : :atom , required : true , doc : &quot;The type of the car&quot; , default : :sedan ] ] @car % Spark.Dsl.Entity { name : :car , describe : &quot;Adds a car&quot; , examples : [ &quot;car :ford, :focus&quot; ] , target : MyApp.Car , args : [ :make , :model ] , schema : @car_schema } @cars % Spark.Dsl.Section { name : :cars , # The DSL constructor will be `cars` describe : &quot;&quot;&quot; Configure what cars are available. More, deeper explanation. Always have a short one liner explanation, an empty line, and then a longer explanation. &quot;&quot;&quot; , entities : [ @car # See `Spark.Dsl.Entity` docs ] , schema : [ default_manufacturer : [ type : :atom , doc : &quot;The default manufacturer&quot; ] ] } use Spark.Dsl.Extension , sections : [ @cars ] end Often, we will need to do complex validation/validate based on the configuration of other resources. Due to the nature of building compile time DSLs, there are many restrictions around that process. To support these complex use cases, extensions can include transformers which can validate/transform the DSL state after all basic sections/entities have been created. See Spark.Dsl.Transformer for more information. Transformers are provided as an option to use , like so: use Spark.Dsl.Extension , sections : [ @cars ] , transformers : [ MyApp.Transformers.ValidateNoOverlappingMakesAndModels ] By default, the generated modules will have names like __MODULE__.SectionName.EntityName , and that could potentially conflict with modules you are defining, so you can specify the module_prefix option, which would allow you to prefix the modules with something like __MODULE__.Dsl , so that the module path generated might be something like __MODULE__.Dsl.SectionName.EntityName , and you could then have the entity struct be __MODULE__.SectionName.EntityName without conflicts. To expose the configuration of your DSL, define functions that use the helpers like get_entities/2 and get_opt/3 . For example: defmodule MyApp.Cars do def cars ( resource ) do Spark.Dsl.Extension . get_entities ( resource , [ :cars ] ) end end MyApp.Cars . cars ( MyResource ) # [%MyApp.Car{...}, %MyApp.Car{...}] See the documentation for Spark.Dsl.Section and Spark.Dsl.Entity for more information","ref":"Spark.Dsl.Extension.html"},{"type":"callback","title":"Spark.Dsl.Extension.add_extensions/0","doc":"","ref":"Spark.Dsl.Extension.html#c:add_extensions/0"},{"type":"function","title":"Spark.Dsl.Extension.doc/2","doc":"See Spark.CheatSheet.doc/2 .","ref":"Spark.Dsl.Extension.html#doc/2"},{"type":"function","title":"Spark.Dsl.Extension.doc_index/2","doc":"See Spark.CheatSheet.doc_index/2 .","ref":"Spark.Dsl.Extension.html#doc_index/2"},{"type":"function","title":"Spark.Dsl.Extension.expand_alias/2","doc":"","ref":"Spark.Dsl.Extension.html#expand_alias/2"},{"type":"function","title":"Spark.Dsl.Extension.expand_alias_no_require/2","doc":"","ref":"Spark.Dsl.Extension.html#expand_alias_no_require/2"},{"type":"function","title":"Spark.Dsl.Extension.expand_literals/3","doc":"","ref":"Spark.Dsl.Extension.html#expand_literals/3"},{"type":"callback","title":"Spark.Dsl.Extension.explain/1","doc":"","ref":"Spark.Dsl.Extension.html#c:explain/1"},{"type":"function","title":"Spark.Dsl.Extension.fetch_opt/4","doc":"","ref":"Spark.Dsl.Extension.html#fetch_opt/4"},{"type":"function","title":"Spark.Dsl.Extension.get_entities/2","doc":"Get the entities configured for a given section","ref":"Spark.Dsl.Extension.html#get_entities/2"},{"type":"function","title":"Spark.Dsl.Extension.get_entity_dsl_patches/2","doc":"","ref":"Spark.Dsl.Extension.html#get_entity_dsl_patches/2"},{"type":"function","title":"Spark.Dsl.Extension.get_opt/5","doc":"Get an option value for a section at a given path. Checks to see if it has been overridden via configuration.","ref":"Spark.Dsl.Extension.html#get_opt/5"},{"type":"function","title":"Spark.Dsl.Extension.get_opt_config/3","doc":"","ref":"Spark.Dsl.Extension.html#get_opt_config/3"},{"type":"function","title":"Spark.Dsl.Extension.get_persisted/3","doc":"Get a value that was persisted while transforming or compiling the resource, e.g :primary_key","ref":"Spark.Dsl.Extension.html#get_persisted/3"},{"type":"macro","title":"Spark.Dsl.Extension.import_mods/1","doc":"","ref":"Spark.Dsl.Extension.html#import_mods/1"},{"type":"callback","title":"Spark.Dsl.Extension.module_imports/0","doc":"","ref":"Spark.Dsl.Extension.html#c:module_imports/0"},{"type":"function","title":"Spark.Dsl.Extension.monotonic_number/1","doc":"","ref":"Spark.Dsl.Extension.html#monotonic_number/1"},{"type":"function","title":"Spark.Dsl.Extension.run_transformers/4","doc":"","ref":"Spark.Dsl.Extension.html#run_transformers/4"},{"type":"callback","title":"Spark.Dsl.Extension.sections/0","doc":"","ref":"Spark.Dsl.Extension.html#c:sections/0"},{"type":"macro","title":"Spark.Dsl.Extension.set_entity_opt/4","doc":"","ref":"Spark.Dsl.Extension.html#set_entity_opt/4"},{"type":"macro","title":"Spark.Dsl.Extension.set_section_opt/6","doc":"","ref":"Spark.Dsl.Extension.html#set_section_opt/6"},{"type":"function","title":"Spark.Dsl.Extension.shuffle_opts_to_end/3","doc":"","ref":"Spark.Dsl.Extension.html#shuffle_opts_to_end/3"},{"type":"function","title":"Spark.Dsl.Extension.spark_function_info/1","doc":"","ref":"Spark.Dsl.Extension.html#spark_function_info/1"},{"type":"callback","title":"Spark.Dsl.Extension.transformers/0","doc":"","ref":"Spark.Dsl.Extension.html#c:transformers/0"},{"type":"macro","title":"Spark.Dsl.Extension.unimport_mods/1","doc":"","ref":"Spark.Dsl.Extension.html#unimport_mods/1"},{"type":"callback","title":"Spark.Dsl.Extension.verifiers/0","doc":"","ref":"Spark.Dsl.Extension.html#c:verifiers/0"},{"type":"type","title":"Spark.Dsl.Extension.t/0","doc":"","ref":"Spark.Dsl.Extension.html#t:t/0"},{"type":"module","title":"Spark.Dsl.Fragment","doc":"Allows splitting up a DSL into multiple modules, potentially organizing large DSLs Use the of option to expression what your fragment is a fragment of. You can add extensions as you would normally to that resource, and they will be added to the parent resource. defmodule MyApp.Resource.Graphql do use Spark.Dsl.Fragment, of: Ash.Resource, extensions: AshGraphql.Resource graphql do ... end end","ref":"Spark.Dsl.Fragment.html"},{"type":"module","title":"Spark.Dsl.Patch.AddEntity","doc":"Supply this when defining an extension to add entity builders to another extension's section. For example @entity % Spark.Dsl.Entity { ... } @dsl_patch % Spark.Dsl.Patch.AddEntity { section_path : [ :foo , :bar ] , entity : @entity } use Spark.Dsl.Extension , dsl_patches : [ @dsl_patch ]","ref":"Spark.Dsl.Patch.AddEntity.html"},{"type":"type","title":"Spark.Dsl.Patch.AddEntity.t/0","doc":"","ref":"Spark.Dsl.Patch.AddEntity.html#t:t/0"},{"type":"module","title":"Spark.Dsl.Section","doc":"Declares a DSL section. A dsl section allows you to organize related configurations. All extensions configure sections, they cannot add DSL builders to the top level. This keeps things organized, and concerns separated. A section may have nested sections, which will be configured the same as other sections. Getting the options/entities of a section is done by providing a path, so you would use the nested path to retrieve that configuration. See Spark.Dsl.Extension.get_entities/2 and Spark.Dsl.Extension.get_opt/4 . A section may have entities, which are constructors that produce instances of structs. For more on entities, see Spark.Dsl.Entity . A section may also have a schema , which is a NimbleOptions schema. Spark will produce builders for those options, so that they may be configured. They are retrieved with Spark.Dsl.Extension.get_opt/4 . To create a section that is available at the top level (i.e not nested inside of its own name), use top_level?: true . Remember, however, that this has no effect on sections nested inside of other sections. For a full example, see Spark.Dsl.Extension .","ref":"Spark.Dsl.Section.html"},{"type":"type","title":"Spark.Dsl.Section.auto_set_fields/0","doc":"","ref":"Spark.Dsl.Section.html#t:auto_set_fields/0"},{"type":"type","title":"Spark.Dsl.Section.describe/0","doc":"User provided documentation. Documentation provided in a Section 's describe field will be included by Spark in any generated documentation that includes the Section .","ref":"Spark.Dsl.Section.html#t:describe/0"},{"type":"type","title":"Spark.Dsl.Section.docs/0","doc":"Internal field. Not set by user.","ref":"Spark.Dsl.Section.html#t:docs/0"},{"type":"type","title":"Spark.Dsl.Section.entities/0","doc":"","ref":"Spark.Dsl.Section.html#t:entities/0"},{"type":"type","title":"Spark.Dsl.Section.examples/0","doc":"","ref":"Spark.Dsl.Section.html#t:examples/0"},{"type":"type","title":"Spark.Dsl.Section.imports/0","doc":"","ref":"Spark.Dsl.Section.html#t:imports/0"},{"type":"type","title":"Spark.Dsl.Section.links/0","doc":"","ref":"Spark.Dsl.Section.html#t:links/0"},{"type":"type","title":"Spark.Dsl.Section.modules/0","doc":"","ref":"Spark.Dsl.Section.html#t:modules/0"},{"type":"type","title":"Spark.Dsl.Section.name/0","doc":"","ref":"Spark.Dsl.Section.html#t:name/0"},{"type":"type","title":"Spark.Dsl.Section.no_depend_modules/0","doc":"","ref":"Spark.Dsl.Section.html#t:no_depend_modules/0"},{"type":"type","title":"Spark.Dsl.Section.patchable?/0","doc":"","ref":"Spark.Dsl.Section.html#t:patchable?/0"},{"type":"type","title":"Spark.Dsl.Section.sections/0","doc":"","ref":"Spark.Dsl.Section.html#t:sections/0"},{"type":"type","title":"Spark.Dsl.Section.snippet/0","doc":"","ref":"Spark.Dsl.Section.html#t:snippet/0"},{"type":"type","title":"Spark.Dsl.Section.t/0","doc":"","ref":"Spark.Dsl.Section.html#t:t/0"},{"type":"type","title":"Spark.Dsl.Section.top_level?/0","doc":"Determines whether a section can be declared directly in a module. When top_level?: true , that Section's DSL can be declared outside of a do block in a module. Example A Section declared with top_level?: true : @my_section % Spark.Dsl.Section { top_level? : true , name : :my_section , schema : [ my_field : [ type : :atom , required : true ] ] } Can be declared like this: defmodule MyDslModule do my_field :value end With top_level?: false , the DSL section would need to be declared explicitly/: defmodule MyDslModule do my_section do my_field :value end end","ref":"Spark.Dsl.Section.html#t:top_level?/0"},{"type":"behaviour","title":"Spark.Dsl.Transformer","doc":"A transformer manipulates and/or validates the entire DSL state of a resource. It's transform/1 takes a map , which is just the values/configurations at each point of the DSL. Don't manipulate it directly, if possible, instead use functions like get_entities/3 and replace_entity/5 to manipulate it. Use the after?/1 and before?/1 callbacks to ensure that your transformer runs either before or after some other transformer. Return true in after_compile/0 to have the transformer run in an after_compile hook, but keep in mind that no modifications to the dsl structure will be retained, so there is no real point in modifying the dsl that you return.","ref":"Spark.Dsl.Transformer.html"},{"type":"function","title":"Spark.Dsl.Transformer.add_entity/4","doc":"","ref":"Spark.Dsl.Transformer.html#add_entity/4"},{"type":"callback","title":"Spark.Dsl.Transformer.after_compile?/0","doc":"","ref":"Spark.Dsl.Transformer.html#c:after_compile?/0"},{"type":"callback","title":"Spark.Dsl.Transformer.after?/1","doc":"","ref":"Spark.Dsl.Transformer.html#c:after?/1"},{"type":"function","title":"Spark.Dsl.Transformer.async_compile/2","doc":"Runs the function in an async compiler. Use this for compiling new modules and having them compiled efficiently asynchronously.","ref":"Spark.Dsl.Transformer.html#async_compile/2"},{"type":"callback","title":"Spark.Dsl.Transformer.before?/1","doc":"","ref":"Spark.Dsl.Transformer.html#c:before?/1"},{"type":"function","title":"Spark.Dsl.Transformer.build_entity!/4","doc":"","ref":"Spark.Dsl.Transformer.html#build_entity!/4"},{"type":"function","title":"Spark.Dsl.Transformer.build_entity/4","doc":"","ref":"Spark.Dsl.Transformer.html#build_entity/4"},{"type":"function","title":"Spark.Dsl.Transformer.eval/3","doc":"Add a quoted expression to be evaluated in the DSL module's context. Use this extremely sparingly . It should almost never be necessary, unless building certain extensions that require the module in question to define a given function. What you likely want is either one of the DSL introspection functions, like Spark.Dsl.Extension.get_entities/2 or Spark.Dsl.Extension.get_opt/5) . If you simply want to store a custom value that can be retrieved easily, or cache some precomputed information onto the resource, use persist/3 . Provide the dsl state, bindings that should be unquote-able, and the quoted block to evaluate in the module. For example, if we wanted to support a resource.primary_key() function that would return the primary key (this is unnecessary, just an example), we might do this: fields = the_primary_key_fields dsl_state = Transformer . eval ( dsl_state , [ fields : fields ] , quote do def primary_key ( ) do unquote ( fields ) end end )","ref":"Spark.Dsl.Transformer.html#eval/3"},{"type":"function","title":"Spark.Dsl.Transformer.fetch_option/3","doc":"","ref":"Spark.Dsl.Transformer.html#fetch_option/3"},{"type":"function","title":"Spark.Dsl.Transformer.get_entities/2","doc":"","ref":"Spark.Dsl.Transformer.html#get_entities/2"},{"type":"function","title":"Spark.Dsl.Transformer.get_option/4","doc":"","ref":"Spark.Dsl.Transformer.html#get_option/4"},{"type":"function","title":"Spark.Dsl.Transformer.get_persisted/3","doc":"","ref":"Spark.Dsl.Transformer.html#get_persisted/3"},{"type":"function","title":"Spark.Dsl.Transformer.persist/3","doc":"Saves a value into the dsl config with the given key. This can be used to precompute some information and cache it onto the resource, or simply store a computed value. It can later be retrieved with Spark.Dsl.Extension.get_persisted/3 .","ref":"Spark.Dsl.Transformer.html#persist/3"},{"type":"function","title":"Spark.Dsl.Transformer.remove_entity/3","doc":"","ref":"Spark.Dsl.Transformer.html#remove_entity/3"},{"type":"function","title":"Spark.Dsl.Transformer.replace_entity/4","doc":"","ref":"Spark.Dsl.Transformer.html#replace_entity/4"},{"type":"function","title":"Spark.Dsl.Transformer.set_option/4","doc":"","ref":"Spark.Dsl.Transformer.html#set_option/4"},{"type":"function","title":"Spark.Dsl.Transformer.sort/1","doc":"","ref":"Spark.Dsl.Transformer.html#sort/1"},{"type":"callback","title":"Spark.Dsl.Transformer.transform/1","doc":"","ref":"Spark.Dsl.Transformer.html#c:transform/1"},{"type":"behaviour","title":"Spark.Dsl.Verifier","doc":"A verifier gets the dsl state and can return :ok or :error . In a verifier, you can reference and depend on other modules without causing compile time dependencies.","ref":"Spark.Dsl.Verifier.html"},{"type":"function","title":"Spark.Dsl.Verifier.fetch_option/3","doc":"See Spark.Dsl.Transformer.fetch_option/3 .","ref":"Spark.Dsl.Verifier.html#fetch_option/3"},{"type":"function","title":"Spark.Dsl.Verifier.get_entities/2","doc":"See Spark.Dsl.Transformer.get_entities/2 .","ref":"Spark.Dsl.Verifier.html#get_entities/2"},{"type":"function","title":"Spark.Dsl.Verifier.get_option/3","doc":"See Spark.Dsl.Transformer.get_option/3 .","ref":"Spark.Dsl.Verifier.html#get_option/3"},{"type":"function","title":"Spark.Dsl.Verifier.get_option/4","doc":"See Spark.Dsl.Transformer.get_option/4 .","ref":"Spark.Dsl.Verifier.html#get_option/4"},{"type":"function","title":"Spark.Dsl.Verifier.get_persisted/2","doc":"See Spark.Dsl.Transformer.get_persisted/2 .","ref":"Spark.Dsl.Verifier.html#get_persisted/2"},{"type":"function","title":"Spark.Dsl.Verifier.get_persisted/3","doc":"See Spark.Dsl.Transformer.get_persisted/3 .","ref":"Spark.Dsl.Verifier.html#get_persisted/3"},{"type":"callback","title":"Spark.Dsl.Verifier.verify/1","doc":"","ref":"Spark.Dsl.Verifier.html#c:verify/1"},{"type":"module","title":"Spark.Dsl.Verifiers.VerifyEntityUniqueness","doc":"Verifies that each entity that has an identifier is unique at each path.","ref":"Spark.Dsl.Verifiers.VerifyEntityUniqueness.html"},{"type":"function","title":"Spark.Dsl.Verifiers.VerifyEntityUniqueness.verify/1","doc":"Callback implementation for Spark.Dsl.Verifier.verify/1 .","ref":"Spark.Dsl.Verifiers.VerifyEntityUniqueness.html#verify/1"},{"type":"module","title":"Spark.OptionsHelpers","doc":"Helpers for working with our superset of nimble options. Additional Types Spark provides the following additional option types for use when building DSLs or validating options. {:one_of, values} - maps to {:in, values} for backwards compatibility {:tagged_tuple, tag, inner_type} - maps to {tag, type} {:spark_behaviour, behaviour} - expects a module that implements the given behaviour, and can be specified with options, i.e mod or {mod, [opt: :val]} {:spark_behaviour, behaviour, builtin_module} - Same as the above, but also accepts a builtin_module . The builtin_module is used to provide additional options for the elixir_sense plugin. {:spark_function_behaviour, behaviour, {function_mod, arity}} - expects a module that implements the given behaviour, and can be specified with options, i.e mod or {mod, [opt: :val]} , that also has a special module that supports being provided an anonymous function or MFA as the :fun option. {:spark_function_behaviour, behaviour, builtin_module, {function_mod, arity}} - Same as the above, but also accepts a builtin_module . The builtin_module is used to provide additional options for the elixir_sense plugin. {:behaviour, behaviour} - expects a module that implements a given behaviour. {:spark, dsl_module} - expects a module that is a Spark.Dsl {:mfa_or_fun, arity} - expects a function or MFA of a corresponding arity. {:spark_type, module, builtin_function} - a behaviour that defines builtin_function/0 that returns a list of atoms that map to built in variations of that thing. {:spark_type, module, builtin_function, templates} - same as the above, but includes additional templates for elixir_sense autocomplete :fun - a function of any arity :literal -&gt; any literal value. Maps to :any , but is used for documentation. {:literal, value} -&gt; exactly the value specified. :quoted -&gt; retains the quoted value of the code provided to the option {:wrap_list, type} -&gt; Allows a single value or a list of values.","ref":"Spark.OptionsHelpers.html"},{"type":"function","title":"Spark.OptionsHelpers.append_doc!/3","doc":"","ref":"Spark.OptionsHelpers.html#append_doc!/3"},{"type":"function","title":"Spark.OptionsHelpers.docs/1","doc":"Creates markdown documentation for a given schema.","ref":"Spark.OptionsHelpers.html#docs/1"},{"type":"function","title":"Spark.OptionsHelpers.fun/1","doc":"","ref":"Spark.OptionsHelpers.html#fun/1"},{"type":"function","title":"Spark.OptionsHelpers.list_of_atoms/1","doc":"","ref":"Spark.OptionsHelpers.html#list_of_atoms/1"},{"type":"function","title":"Spark.OptionsHelpers.literal_value/2","doc":"","ref":"Spark.OptionsHelpers.html#literal_value/2"},{"type":"function","title":"Spark.OptionsHelpers.make_optional!/2","doc":"","ref":"Spark.OptionsHelpers.html#make_optional!/2"},{"type":"function","title":"Spark.OptionsHelpers.make_required!/2","doc":"","ref":"Spark.OptionsHelpers.html#make_required!/2"},{"type":"function","title":"Spark.OptionsHelpers.map/1","doc":"","ref":"Spark.OptionsHelpers.html#map/1"},{"type":"function","title":"Spark.OptionsHelpers.merge_schemas/3","doc":"Merges two schemas, and sets the subsection option on all options on the right side.","ref":"Spark.OptionsHelpers.html#merge_schemas/3"},{"type":"function","title":"Spark.OptionsHelpers.mfa_or_fun/2","doc":"","ref":"Spark.OptionsHelpers.html#mfa_or_fun/2"},{"type":"function","title":"Spark.OptionsHelpers.module_and_opts/1","doc":"","ref":"Spark.OptionsHelpers.html#module_and_opts/1"},{"type":"function","title":"Spark.OptionsHelpers.sanitize_schema/1","doc":"","ref":"Spark.OptionsHelpers.html#sanitize_schema/1"},{"type":"function","title":"Spark.OptionsHelpers.set_default!/3","doc":"","ref":"Spark.OptionsHelpers.html#set_default!/3"},{"type":"function","title":"Spark.OptionsHelpers.set_type!/3","doc":"","ref":"Spark.OptionsHelpers.html#set_type!/3"},{"type":"function","title":"Spark.OptionsHelpers.spark_behaviour/2","doc":"","ref":"Spark.OptionsHelpers.html#spark_behaviour/2"},{"type":"function","title":"Spark.OptionsHelpers.spark_function_behaviour/3","doc":"","ref":"Spark.OptionsHelpers.html#spark_function_behaviour/3"},{"type":"function","title":"Spark.OptionsHelpers.struct/2","doc":"","ref":"Spark.OptionsHelpers.html#struct/2"},{"type":"function","title":"Spark.OptionsHelpers.tagged_tuple/4","doc":"","ref":"Spark.OptionsHelpers.html#tagged_tuple/4"},{"type":"function","title":"Spark.OptionsHelpers.validate!/2","doc":"Sanitizes the option schema and validates with NimbleOptions.","ref":"Spark.OptionsHelpers.html#validate!/2"},{"type":"function","title":"Spark.OptionsHelpers.validate/2","doc":"Sanitizes the option schema and validates with NimbleOptions.","ref":"Spark.OptionsHelpers.html#validate/2"},{"type":"function","title":"Spark.OptionsHelpers.wrap_list/3","doc":"","ref":"Spark.OptionsHelpers.html#wrap_list/3"},{"type":"type","title":"Spark.OptionsHelpers.nimble_schema/0","doc":"The schema type, as defined by the NimbleOptions docs .","ref":"Spark.OptionsHelpers.html#t:nimble_schema/0"},{"type":"type","title":"Spark.OptionsHelpers.nimble_types/0","doc":"The types as specified by the NimbleOptions docs .","ref":"Spark.OptionsHelpers.html#t:nimble_types/0"},{"type":"type","title":"Spark.OptionsHelpers.schema/0","doc":"The actual schema type, as used in sections and entities.","ref":"Spark.OptionsHelpers.html#t:schema/0"},{"type":"type","title":"Spark.OptionsHelpers.type/0","doc":"Spark provides additional types on top of those provided by NimbleOptions .","ref":"Spark.OptionsHelpers.html#t:type/0"},{"type":"exception","title":"Spark.Error.DslError","doc":"Used when a DSL is incorrectly configured.","ref":"Spark.Error.DslError.html"},{"type":"type","title":"Spark.Error.DslError.t/0","doc":"","ref":"Spark.Error.DslError.html#t:t/0"},{"type":"task","title":"Mix.Tasks.Spark.CheatSheets","doc":"Creates a set of cheat sheets for each Extension provided","ref":"Mix.Tasks.Spark.CheatSheets.html"},{"type":"function","title":"Mix.Tasks.Spark.CheatSheets.run/1","doc":"Callback implementation for Mix.Task.run/1 .","ref":"Mix.Tasks.Spark.CheatSheets.html#run/1"},{"type":"task","title":"Mix.Tasks.Spark.CheatSheetsInSearch","doc":"Includes generated cheat sheets in the search bar","ref":"Mix.Tasks.Spark.CheatSheetsInSearch.html"},{"type":"function","title":"Mix.Tasks.Spark.CheatSheetsInSearch.run/1","doc":"Callback implementation for Mix.Task.run/1 .","ref":"Mix.Tasks.Spark.CheatSheetsInSearch.html#run/1"},{"type":"task","title":"Mix.Tasks.Spark.Formatter","doc":"Manages a variable called spark_locals_without_parens in the .formatter.exs from a list of DSL extensions.","ref":"Mix.Tasks.Spark.Formatter.html"},{"type":"function","title":"Mix.Tasks.Spark.Formatter.all_entity_builders_everywhere/3","doc":"","ref":"Mix.Tasks.Spark.Formatter.html#all_entity_builders_everywhere/3"},{"type":"function","title":"Mix.Tasks.Spark.Formatter.run/1","doc":"Callback implementation for Mix.Task.run/1 .","ref":"Mix.Tasks.Spark.Formatter.html#run/1"},{"type":"module","title":"Spark","doc":"Documentation for Spark .","ref":"Spark.html"},{"type":"function","title":"Spark.extensions/1","doc":"Returns the extensions a given DSL uses","ref":"Spark.html#extensions/1"},{"type":"function","title":"Spark.implements_behaviour?/2","doc":"Returns true if the module implements the specified behavior","ref":"Spark.html#implements_behaviour?/2"},{"type":"function","title":"Spark.otp_app/1","doc":"Returns the configured otp_app of a given DSL instance","ref":"Spark.html#otp_app/1"},{"type":"function","title":"Spark.sparks/2","doc":"Returns all modules that implement the specified behaviour for a given otp_app. Should only be called at runtime, not at compile time, as it will have inconsistent results at compile time.","ref":"Spark.html#sparks/2"},{"type":"module","title":"Spark.CheatSheet","doc":"Tools to generate cheat sheets for spark DSLs","ref":"Spark.CheatSheet.html"},{"type":"function","title":"Spark.CheatSheet.cheat_sheet/1","doc":"Generate a cheat sheet for a given DSL","ref":"Spark.CheatSheet.html#cheat_sheet/1"},{"type":"function","title":"Spark.CheatSheet.doc/2","doc":"Generate a markdown bullet list documentation for a list of sections","ref":"Spark.CheatSheet.html#doc/2"},{"type":"function","title":"Spark.CheatSheet.doc_index/3","doc":"Generate a table of contents for a list of sections","ref":"Spark.CheatSheet.html#doc_index/3"},{"type":"function","title":"Spark.CheatSheet.section_cheat_sheet/2","doc":"","ref":"Spark.CheatSheet.html#section_cheat_sheet/2"},{"type":"module","title":"Spark.CodeHelpers","doc":"Helpers for meta programming around code and code snippets","ref":"Spark.CodeHelpers.html"},{"type":"function","title":"Spark.CodeHelpers.code_identifier/1","doc":"Given a section of Elixir AST, generate a hash of the code to help with generating unique names.","ref":"Spark.CodeHelpers.html#code_identifier/1"},{"type":"function","title":"Spark.CodeHelpers.lift_functions/3","doc":"Lift anonymous and captured functions. Acts as an AST transformer to allow these kinds of functions to be added in the AST: In the case of captured functions, it ensures they are all captured remote functions (ie calls with both the module and function name present) - this often requires the definition of a new public function on the target module. In the case of anonymous functions, it converts them into a new public function on the module and returns a (remote) function capture much like that of above.","ref":"Spark.CodeHelpers.html#lift_functions/3"},{"type":"module","title":"Spark.Formatter","doc":"Formats Spark modules. Currently, it is very simple, and will only reorder the outermost sections according to some rules. Plugin Include the plugin into your .formatter.exs like so plugins: [Spark.Formatter] . If no configuration is provided, it will sort all top level DSL sections alphabetically . Section Order To provide a custom section order, add configuration to your app, for example: config :spark , :formatter , remove_parens? : true , &quot;Ash.Resource&quot; : [ section_order : [ :resource , :postgres , :attributes , :relationships , :aggregates , :calculations ] ] , &quot;MyApp.Resource&quot; : [ # Use this if you use a module that is not the spark DSL itself. # For example, you might have a &quot;base&quot; that you use instead that sets some simple defaults. # This tells us what the actual thing is so we know what extensions are included automatically. type : Ash.Resource , # Tell us what extensions might be added under the hood extensions : [ MyApp.ResourceExtension ] , section_order : [ ... ] ] Any sections found that aren't in that list will be left in the order that they were in, the sections in the list will be sorted &quot;around&quot; those sections. E.g the following list: [:code_interface, :attributes] can be interpreted as &quot;ensure that code_interface comes before attributes, and don't change the rest&quot;.","ref":"Spark.Formatter.html"},{"type":"function","title":"Spark.Formatter.features/1","doc":"Callback implementation for Mix.Tasks.Format.features/1 .","ref":"Spark.Formatter.html#features/1"},{"type":"function","title":"Spark.Formatter.format/2","doc":"Callback implementation for Mix.Tasks.Format.format/2 .","ref":"Spark.Formatter.html#format/2"},{"type":"function","title":"Spark.Formatter.prewalk/2","doc":"Copy of Macro.prewalk/2 w/ a branch accumulator","ref":"Spark.Formatter.html#prewalk/2"},{"type":"function","title":"Spark.Formatter.prewalk/4","doc":"Copy of Macro.prewalk/3 w/ a branch accumulator","ref":"Spark.Formatter.html#prewalk/4"},{"type":"function","title":"Spark.Formatter.traverse/5","doc":"A copy of the corresponding Macro.traverse function that has a separate accumulator that only goes down each branch, only for pre","ref":"Spark.Formatter.html#traverse/5"},{"type":"module","title":"Spark.InfoGenerator","doc":"Used to dynamically generate configuration functions for Spark extensions based on their DSL. Usage defmodule MyConfig do use Spark.InfoGenerator , extension : MyDslExtension , sections : [ :my_section ] end","ref":"Spark.InfoGenerator.html"},{"type":"macro","title":"Spark.InfoGenerator.generate_config_functions/2","doc":"Given an extension and a list of DSL sections generate individual config functions for each option.","ref":"Spark.InfoGenerator.html#generate_config_functions/2"},{"type":"macro","title":"Spark.InfoGenerator.generate_entity_functions/2","doc":"Given an extension and a list of DSL sections, generate an entities function which returns a list of entities.","ref":"Spark.InfoGenerator.html#generate_entity_functions/2"},{"type":"macro","title":"Spark.InfoGenerator.generate_options_functions/2","doc":"Given an extension and a list of DSL sections, generate an options function which returns a map of all configured options for a resource (including defaults).","ref":"Spark.InfoGenerator.html#generate_options_functions/2"},{"type":"function","title":"Spark.InfoGenerator.spec_for_type/2","doc":"","ref":"Spark.InfoGenerator.html#spec_for_type/2"},{"type":"type","title":"Spark.InfoGenerator.options/0","doc":"","ref":"Spark.InfoGenerator.html#t:options/0"},{"type":"module","title":"Spark.Types","doc":"Helpers for spark types","ref":"Spark.Types.html"},{"type":"function","title":"Spark.Types.doc_type/1","doc":"","ref":"Spark.Types.html#doc_type/1"},{"type":"extras","title":"Writing Extensions","doc":"Writing extensions generally involves three main components.","ref":"writing-extensions.html"},{"type":"extras","title":"Writing Extensions - The DSL declaration","doc":"The DSL is declared as a series of Spark.Dsl.Section , which can contain Spark.Dsl.Entity and further Spark.Dsl.Section structs. See Spark.Dsl.Section and Spark.Dsl.Entity for more information.","ref":"writing-extensions.html#the-dsl-declaration"},{"type":"extras","title":"Writing Extensions - Transformers","doc":"Extension writing gets a bit more complicated when you get into the world of transformers, but this is also where a lot of the power is. Each transformer can declare other transformers it must go before or after, and then is given the opportunity to modify the entirety of the DSL it is extending up to that point. This allows extensions to make rich modifications to the structure in question. See Spark.Dsl.Transformer for more information","ref":"writing-extensions.html#transformers"},{"type":"extras","title":"Writing Extensions - Introspection","doc":"Use functions in Spark.Dsl.Extension to retrieve the stored values from the DSL and expose them in a module. The convention is to place functions for something like MyApp.MyExtension in MyApp.MyExtension.Info . Using introspection functions like this allows for a richer introspection API (i.e not just getting and retrieving raw values), and it also allows us to add type specs and documentation, which is helpful when working generically. I.e module_as_variable.table() can't be known by dialyzer, whereas Extension.table(module) can be.","ref":"writing-extensions.html#introspection"},{"type":"extras","title":"Get Started With Spark","doc":"Spark helps you build powerful and well documented DSLs that come with useful tooling out of the box. DSLs are declared using simple structs, and every DSL has the ability to be extended by the end user. Spark powers all of the DSLs in Ash Framework. What you get for your DSL when you implement it with Spark: Extensibility. Anyone can write extensions for your DSL. Autocomplete and in-line documentation: An elixir_sense plugin that &quot;Just Works&quot; for any DSL implemented with Spark. Tools to generate documentation for your DSL automatically. A mix task to add every part of the DSL to the locals_without_parens of your library automatically. This library has only recently been extracted out from Ash core, so there is still work to be done to document and test it in isolation. See the module documentation for the most up-to-date information.","ref":"get-started-with-spark.html"},{"type":"extras","title":"Get Started With Spark - Dependency","doc":"{:spark, &quot;~&gt; 1.1.32&quot;}","ref":"get-started-with-spark.html#dependency"}]